--[[
This Source Code Form is subject to the terms of 
the Mozilla Public License, v. 2.0. If a copy of 
the MPL was not distributed with this file, You 
can obtain one at https://mozilla.org/MPL/2.0/.

Copyright © 2022 The Sand Reckoner
Gmail: sandreckoner1063
]]

local addon_name, addon_table = ...
local EnhancedModelScene = addon_table.self
local EMS = EnhancedModelScene

if not EMS then return end

local Script = {}
EnhancedModelScene.Script = Script


function Script:Loop(script)
	for line in script:gmatch("[^\n]+") do
		-- FIXME: check for a comment character, and maybe add a rem command
		EMS:DispatchCommand(line)
	end
	
	print("done executing script")
end


function Script:Run(script)
	-- use coroutine.create rather than wrap, because we can 
	-- probably continue to the next line after an error
	-- Or maybe implement error catching in the loop function?
	local thread = coroutine.create(function() Script:Loop(script) end)
	
	function sleep_handler()
		local status, sleep_duration = coroutine.resume(thread)
		
		if status and sleep_duration then
			C_Timer.After(sleep_duration, sleep_handler)
		end
	end
	
	sleep_handler()
end


Script.test_script = [[
walk 1 1
sleep 1
turn 1 1
sleep 2
turn 1 0
walk 1 0
anim 1 60
sleep 2
anim 1 0
]]